import time
import torch
import torch.nn.utils as utils

import pytorch_lightning as pl
from pytorch_custom_utils import get_adam_optimizer

from beartype import beartype

from .ema import EMA
from .optimizer_scheduler import (
    get_cosine_schedule_with_warmup,
)

from collections import namedtuple

ModelOuput = namedtuple(
    "ModelOuput", ["loss", "report", "output"], defaults=[None, None, None]
)


class ModelWrapper(pl.LightningModule):
    @beartype
    def __init__(
        self,
        *,
        model: torch.nn.Module,
        use_ema: bool = False,
        scheduler_kwargs: dict = dict(),
        optimizer_kwargs: dict = dict(),
        forward_kwargs: dict = dict(),
        **kwargs,
    ):
        super().__init__()
        self.model = model
        self.forward_kwargs = forward_kwargs

        self.optimizer = optimizer_kwargs.pop("optimizer", None)
        self.schedueler = scheduler_kwargs.pop("scheduler", None)

        self.optimizer_kwargs = optimizer_kwargs
        self.scheduler_kwargs = scheduler_kwargs

        self.max_grad_norm = optimizer_kwargs.get("max_grad_norm", 1.0)
        self.total_norm = 0

        self.wandb_id = None
        self.start_step = 0
        self.start_epoch = 0

        self.use_ema = use_ema
        if self.use_ema:
            self.ema_model = EMA(self.model)

        print("Unuserd kwargs:", kwargs)

    def configure_optimizers(self):
        optimizer = (
            self.optimizer
            if self.optimizer is not None
            else get_adam_optimizer(self.model.parameters(), **self.optimizer_kwargs)
        )
        scheduler = (
            self.schedueler
            if self.schedueler is not None
            else get_cosine_schedule_with_warmup(optimizer, **self.scheduler_kwargs)
        )
        return [optimizer], [scheduler]

    def optimizer_step(self, *args, **kwargs):
        optimizer = self.optimizers()
        all_params = [
            p
            for param_group in optimizer.param_groups
            for p in param_group["params"]
            if p.requires_grad
        ]
        if self.max_grad_norm is not None:
            self.total_norm = utils.clip_grad_norm_(all_params, self.max_grad_norm)
        else:
            self.total_norm = utils.clip_grad_norm_(all_params, float("inf"))
        super().optimizer_step(*args, **kwargs)

    def forward(self, x):
        output = self.model(**x, **self.forward_kwargs)
        return ModelOuput(**output)

    def training_step(self, batch, batch_idx):
        fwd_out = self(batch)
        loss = fwd_out.loss
        report = fwd_out.report

        for k, v in report.items():
            self.log("training/" + k, v, logger=True, sync_dist=True)

        if self.use_ema and self.global_rank == 0:
            self.ema_model.step(self.model)
        return loss

    def validation_step(self, batch, batch_idx):
        fwd_out = self(batch)
        loss = fwd_out.loss
        report = fwd_out.report
        for k, v in report.items():
            self.log("validation/" + k, v, logger=True, sync_dist=True)
        return loss

    def on_save_checkpoint(self, checkpoint):
        """Manually save additional metrics."""
        checkpoint["metrics"] = self.trainer.callback_metrics
        if self.use_ema:
            checkpoint["ema_state_dict"] = self.ema_model.state_dict()
        checkpoint["wandb_id"] = self.trainer.logger.experiment.id

    def on_load_checkpoint(self, checkpoint):
        super().on_load_checkpoint(checkpoint)
        self.trainer.callback_metrics.update(checkpoint.get("metrics", {}))

        self.start_step = checkpoint.get("global_step", 0)
        self.start_epoch = checkpoint.get("epoch", 0)
        if self.use_ema:
            if "ema_state_dict" in checkpoint:
                self.ema_model.load_state_dict(checkpoint["ema_state_dict"])
                print("EMA state dict found in checkpoint.")
            else:
                print("No EMA state dict found in checkpoint.")



import pytorch_lightning as pl
from torch.utils.data import DataLoader, Dataset
from typeguard import typechecked

from .data_loader import DictDataLoader


@typechecked
class SharedDataModule(pl.LightningDataModule):
    def __init__(
        self,
        dataset_class: type[Dataset],
        dataloader_class: type[DataLoader] = DictDataLoader,
        training_kwargs: dict = dict(),
        validation_kwargs: dict = dict(),
        dataloader_kwargs: dict = dict(),
    ):
        super().__init__()
        self.dataset_class = dataset_class
        self.dataloader_class = dataloader_class
        self.train_kwargs = training_kwargs
        self.val_kwargs = validation_kwargs
        self.dl_kwargs = dataloader_kwargs

        self.train_dataset = None
        self.val_dataset = None

    def setup(self, stage=None):
        if self.train_dataset is None:
            self.train_dataset = self.dataset_class(**self.train_kwargs)

        if self.val_kwargs is not None:
            if self.val_dataset is None:
                self.val_dataset = self.dataset_class(**self.val_kwargs)

    def train_dataloader(self):
        return self.dataloader_class(self.train_dataset, **self.dl_kwargs, shuffle=True)

    def val_dataloader(self):
        if self.val_dataset is None:
            return None
        return self.dataloader_class(self.val_dataset, **self.dl_kwargs, shuffle=False)



import os
import yaml
import pytorch_lightning as pl
import torch

torch.backends.cudnn.benchmark = True
from pytorch_lightning.loggers import WandbLogger

from data_utils.data_module import SharedDataModule, DictDataLoader
from trainer_utils.model_wrapper import ModelWrapper
from trainer_utils.callbacks import (
    SaveCheckpoint,
    LogLearningRate,
    LogGradient,
    LogETL,
)


class DummyDataset(torch.utils.data.Dataset):
    def __init__(self, **kwargs):
        self.length = kwargs.get("length", 100)
        super().__init__()

    def __len__(self):
        return self.length

    def __getitem__(self, idx):
        vertices = torch.randn(32, 2)
        segments = torch.randint(0, 32, (10, 2))
        return {
            "vertices": vertices,
            "segments": segments,
        }


class DummyModel(torch.nn.Module):
    def __init__(self, **kwargs):
        super().__init__()
        self.linear = torch.nn.Linear(2, 200000)

    def forward(self, vertices, **kwargs):
        x = self.linear(vertices)
        loss = torch.mean(x)
        report = {"loss": loss}

        return {"loss": loss, "report": report, "output": x}


if __name__ == "__main__":
    all_data_kwargs = yaml.safe_load(open("configs/data.yaml", "r"))
    model_kwargs = yaml.safe_load(open("configs/model.yaml", "r"))
    trainer_kwargs = yaml.safe_load(open("configs/trainer.yaml", "r"))

    data_kwargs = all_data_kwargs.get("data", dict())
    train_data_kwargs = data_kwargs.get("general", dict()).copy()
    train_data_kwargs.update(data_kwargs.get("train", dict()))

    data_kwargs = all_data_kwargs.get("data", dict())
    val_data_kwargs = data_kwargs.get("general", dict()).copy()
    val_data_kwargs.update(data_kwargs.get("validataion", dict()))

    datamodule = SharedDataModule(
        dataloader_class=DictDataLoader,
        dataset_class=DummyDataset,
        training_kwargs=train_data_kwargs,
        validation_kwargs=val_data_kwargs,
        dataloader_kwargs=all_data_kwargs.get("dataloader", dict()),
    )

    model = DummyModel(**model_kwargs)
    wrapped_model = ModelWrapper(model=model, **trainer_kwargs.get("wrapper", dict()))

    ckpt_path = trainer_kwargs.get("ckpt_path", None)
    if ckpt_path is not None and os.path.exists(ckpt_path):
        torch.load(ckpt_path, weights_only=True)
    else:
        print(f"Checkpoint not found at {ckpt_path}. Starting from scratch.")
        ckpt_path = None

    wandb_logger = WandbLogger(
        **trainer_kwargs.get("wandb", dict()), id=wrapped_model.wandb_id
    )
    callbacks = [SaveCheckpoint(), LogETL(), LogGradient(), LogLearningRate()]

    trainer = pl.Trainer(
        **trainer_kwargs.get("trainer", dict()),
        logger=wandb_logger,
        callbacks=callbacks,
        log_every_n_steps=1,
    )

    trainer.fit(wrapped_model, datamodule=datamodule, ckpt_path=ckpt_path)
